"""
[Mar. 2021] 2018 Chicago Salaries - Mina Pecheux

A quick ML study of open-source data to show the importance of the training
dataset in machine learning models predictions.
===========================================================================
STEP 1: DATA PREPARATION
------------------------
Transform all hourly rates to annual salaries + add the "gender"
categorical feature based on the firstnames.
"""

import re
import pandas as pd

FIRSTNAME_REGEX = r'.+,\s+(\w+)'

def get_annual_salary(data):
  """Gets the annual salary for all rows in the given dataset. For rows that
  already have an annual salary, we simply remove the $ dollar sign; for rows
  that have an hourly rate, we also convert it to an annual salary.

  Parameters
  ----------
  data: pd.DataFrame
    Dataset to read rows from.

  Returns
  -------
  pd.Series
    Values for the annual salary column - one for each row in the given
    dataset.
  """
  return pd.Series(
    float(row._7[1:]) if isinstance(row._7, str)
    else float(row._8[1:]) * row._6 * 52
    for row in data.itertuples()
  )

def get_gender(row, male_firstnames, female_firstnames):
  """Extracts the gender of the person in the given record based on their
  firstname. The firstname is matched to either the list of male or female
  firstnames... if none matches, the gender is set to None.

  Parameters
  ----------
  row: pd.Series
    Row to add the "gender" feature to.
  male_firstnames: list(str)
    List of known male firstnames to search a match in.
  female_firstnames: list(str)
    List of known female firstnames to search a match in.

  Returns
  -------
  str, None
    Gender of the person in the record.
  """
  firstname = re.search(FIRSTNAME_REGEX, row.Name).group(1).lower()
  if firstname in male_firstnames:
    return 'male'
  if firstname in female_firstnames:
    return 'female'
  return None


if __name__ == '__main__':
  # 1. load the datasets and util firstname lists
  print('1. Loading datasets...')
  # . from: https://www.kaggle.com/gfennema/2018-chicago-employees-salaries
  df = pd.read_csv('datasets/chicago_employees.csv')
  # . from: https://media.geeksforgeeks.org/wp-content/uploads/male.txt
  with open('datasets/male_firstnames.txt') as FILE:
    male_firstnames = [n.lower().strip() for n in FILE.readlines()]
  # . from: https://media.geeksforgeeks.org/wp-content/uploads/female.txt
  with open('datasets/female_firstnames.txt') as FILE:
    female_firstnames = [n.lower().strip() for n in FILE.readlines()]

  # 2. convert hourly rates to annual salaries (and drop rows missing
  # a value after conversion - there shouldn't be any but this is a
  # failsafe)
  print('2. Computing annual salaries')
  df['Annual Salary'] = get_annual_salary(df)
  df.dropna(subset=['Annual Salary'], inplace=True)

  # 3. add the "gender" categorical feature by matching the firstname to
  # either the list of male or female firstnames
  # (if none match, then drop the row)
  print('3. Adding "gender" feature')
  df['Gender'] = pd.Series(
    get_gender(row, male_firstnames, female_firstnames) for row in df.itertuples()
  )
  df.dropna(subset=['Gender'], inplace=True)

  # 4. export the prepared dataset
  print('4. Exporting the transformed dataset')
  df.to_csv('datasets/chicago_employees_prepared.csv', index=False)
