"""
[Mar. 2021] 2018 Chicago Salaries - Mina Pecheux

A quick ML study of open-source data to show the importance of the training
dataset in machine learning models predictions.
===========================================================================
STEP 3: MODELS TRAINING
-----------------------
Train either a linear regression (lr) or a random forest (rf) and predict
annual salaries for some random samples. Optionnally, artificially
"rebalance" the dataset to give men and women with similar qualifications
equal salaries.
"""

import argparse
from random import uniform

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from tabulate import tabulate

REGRESSORS = {
  'lr': LinearRegression,
  'rf': RandomForestRegressor
}

FEATURE_IMPORTANCE_KEY = {
  'lr': 'coef_',
  'rf': 'feature_importances_'
}

if __name__ == '__main__':
  # 0. get command line arguments
  parser = argparse.ArgumentParser()
  parser.add_argument('-m', '--model', type=str, choices=['lr', 'rf'], required=True,
    help='Model to train; can be "lr" (linear regression) or "rf" (random forest)')
  parser.add_argument('--rebalance', action='store_true')
  args = parser.parse_args()

  model = args.model
  rebalance = args.rebalance
  print('0. Starting training with the following parameters:')
  print('- model = {}'.format(model))
  print('- rebalance = {}'.format(rebalance))
  print('')

  # 1. import prepared dataset + separate men/women populations
  print('1. Importing prepared data')
  df = pd.read_csv('datasets/chicago_employees_prepared.csv')

  # 1b. rebalance if needed
  # (copy the data for men, change the gender and apply some randomness,
  # then take a third of the rows to recreate the initial 75%/25% repartition)
  if rebalance:
    print('>> "Rebalancing" dataset')
    only_males = df[df['Gender'] == 'male']
    only_females = pd.DataFrame.copy(only_males)
    only_females['Gender'] = 'female'
    only_females.loc[
      only_females['Full or Part-Time'] == 'F', 'Annual Salary'
    ] = only_females['Annual Salary'] * uniform(0.99, 1.01)
    only_females = only_females.loc[np.random.choice(only_females.index, len(only_males) // 3, replace=False)]
    df = pd.concat([only_males, only_females])
    print('')
  
  # 2. prepare features and target:
  # - convert all strings to categorical features (as integers using
  # the categories codes); e.g.: "POLICE OFFICER" will be transformed
  # into 743
  # - take the annual salary as target
  print('2. Preparing features and target')
  X = pd.DataFrame({
    'Job Titles': df['Job Titles'].astype('category').cat.codes,
    'Department': df['Department'].astype('category').cat.codes,
    'Full or Part-Time': df['Full or Part-Time'].astype('category').cat.codes,
    'Gender': (df['Gender'] == 'male').astype('int')
  })
  y = df['Annual Salary']

  # 3. train the ML model
  print('3. Training the model ({})'.format(model))
  regressor = REGRESSORS[model]()
  regressor.fit(X, y)

  # 4. show the feature importance
  print('4. Getting the feature importances:')
  feature_importances = getattr(regressor, FEATURE_IMPORTANCE_KEY[model])
  feature_importances_table = [
    [column, feature_importances[i]]
    for i, column in enumerate(X.columns.values)
  ]
  print(tabulate(
    feature_importances_table,
    headers=['Feature', 'Importance'],
    numalign='right',
    tablefmt='fancy_grid'
  ))

  # 5. use the model to predict salaries for random samples (with a
  # 75%/25% male/female repartition)
  n_samples_male = 30000
  n_samples_female = 10000
  random_males = np.array([
    # (random job title among the 1,050 available)
    np.random.randint(1050, size=n_samples_male),
    # (random job department among the 34 available)
    np.random.randint(34, size=n_samples_male),
    # (full/part time job is a boolean feature, 0 or 1)
    np.random.randint(2, size=n_samples_male),
    # "male" gender is encoded as 1
    [1] * n_samples_male,
  ]).transpose()
  random_females = np.array([
    np.random.randint(1050, size=n_samples_female),
    np.random.randint(34, size=n_samples_female),
    np.random.randint(2, size=n_samples_female),
    # "female" gender is encoded as 0
    [0] * n_samples_female,
  ]).transpose()

  preds_male = regressor.predict(random_males)
  preds_female = regressor.predict(random_females)
  d = {'male': preds_male, 'female': preds_female}
  preds_by_gender = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in d.items() ]))
  mean_predicted_salaries = preds_by_gender.mean()

  print('')
  print('6. Getting mean salaries for predicted samples (in $/year):')
  print(mean_predicted_salaries)
  print('')

  print('7. Plotting predicted annual salaries repartition by gender')
  preds_by_gender.plot.hist(bins=80, alpha=0.4, color=['blue', 'red'])
  plt.vlines(x=mean_predicted_salaries['male'], ymin=0, ymax=1500, color='blue')
  plt.vlines(x=mean_predicted_salaries['female'], ymin=0, ymax=1500, color='red')
  plt.title('Predicted annual salary repartition by gender\n(model : {})'.format(model))
  plt.xlabel('Predicted annual salary ($)')
  plt.ylabel('Number of employees')
  plt.show()
